#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Initialize blast computation
# Paul Dechamps

AVAILABLE_SOLVERS = ['DART']

def init_blaster(cfg, icfg, iSolverName='DART', task='analysis'):
    """
    Inputs
    ------
    - cfg: Dict with options
        - Re (float): Flow Reynolds number.
        - Minf (float): Freestream Mach number (only used for time step computation).
        - CFL0 (float): Initial CFL number.
        - Verb (int): Verbosity level of the viscous solver.
        - xtrF ([float, float]): Forced transition location [upper, lower].

        - couplIter (int): Maximum number of coupling iterations.
        - couplTol (float): Tolerance to end computation (drag count between 2 iterations).
        - iterPrint (int): Number of iterations between which the coupler outputs its state.
        - resetInv (bool): Resets the inviscid solver at each iteration (True), reuses previous
                            state as initial condition (False).
        Note
        All inputs have default values except Re.
    - icg: Dict with options. Inviscid solver configuration.
    - iSolverName (string): Name of the inviscid solver.

    Outputs
    -------
    dict with keys
        - coupler: Coupler python object to run the viscous-inviscid algorithm.
        - solversAPI: Interface between the solver objects to ensure proper communication.
        - vSolver: blast::Driver object to run the viscous computation.
        - iSolverObjects: Dict containing.
            - All inviscid solver dependencies (depend on iSolverName).

    @todo: - Correctly handle 3D computation parameters.
    """
    # Imports
    from blast.blCoupler import Coupler
    import blast

    if iSolverName not in AVAILABLE_SOLVERS:
        raise RuntimeError('BLASTERAPI: Invalid inviscid solver name', iSolverName)
    elif iSolverName == 'DART':
        from blast.interfaces.dart.blDartInterface import DartInterface as interface

    # Viscous solver
    if 'Re' not in cfg or cfg['Re'] <= 0.:
        raise RuntimeError('BLASTERAPI: Missing or invalid Reynolds number')
    _Re = cfg['Re']

    _Minf = cfg.get('Minf', 0.1) # Used for time step computation only.
    _cfl0 = cfg.get('CFL0', 1.)
    _verbose = cfg.get('Verb', 0)
    _xtrF = cfg.get('xtrF', [-1, -1])
    _span = cfg.get('span', 1.)
    if 'sections' not in cfg:
        raise RuntimeError('BLASTERAPI: Missing sections in the configuration')
    _nSections = len(cfg['sections'])

    if _Minf <= 0:
        raise RuntimeError('BLASTERAPI: Invalid Mach number', _Minf)
    if _cfl0 <= 0:
        raise RuntimeError('BLASTERAPI: Invalid CFL number', _cfl0)
    for xtr in _xtrF:
        if not(0 <= xtr <= 1) and xtr != -1:
            raise RuntimeError("BLASTERAPI: Incorrect forced transition location.", xtr)
    if _span < 0:
        raise RuntimeError('BLASTERAPI: Invalid span', _span)
    if _nSections <= 0:
        raise RuntimeError('BLASTERAPI: Invalid number of sections', _nSections)

    # Coupler
    _couplIter = cfg.get('couplIter', 150)
    _couplTol = cfg.get('couplTol', 1e-4)
    _iterPrint = cfg.get('iterPrint', 1)
    _resetInv = cfg.get('resetInv', False)

    if _couplIter < 0:
        raise RuntimeError('BLASTERAPI: Invalid number of coupling iterations', _couplIter)
    if _couplTol <= 0:
        raise RuntimeError('BLASTERAPI: Invalid coupling tolerance', _couplTol)
    if _iterPrint < 0:
        raise RuntimeError('BLASTERAPI: Invalid iteration print frequency', _iterPrint)

    # Viscous solver object.
    vSolver = blast.Driver(_Re, _Minf, _cfl0, _nSections, _xtrF[0], _xtrF[1], _span, _verbose =_verbose)
    # Solvers interface.
    solverAPI = interface(icfg, vSolver, cfg)
    # Coupler
    coupler = Coupler(solverAPI, vSolver, _maxCouplIter = _couplIter, _couplTol = _couplTol, _iterPrint = _iterPrint, _resetInv = _resetInv)

    # Sanity check
    if solverAPI.getnDim() == 3:
        if 'span' not in cfg:
            raise RuntimeError('BLASTERAPI: Missing span for 3D computation')
        if cfg['interpolator'] != 'Rbf':
            raise RuntimeError('BLASTERAPI: 3D computation only supports Rbf interpolator')

    # Adjoint objects
    cAdj = None
    if task == 'optimization':
        cAdj = blast.CoupledAdjoint(solverAPI.adjointSolver, vSolver)

    return {'coupler': coupler,
            'isol': solverAPI,
            'vsol': vSolver,
            'adj': cAdj}
