#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2024 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## BLASTER api for openMDAO
# Paul Dechamps

import numpy as np
from fwk.coloring import ccolors
import openmdao.api as om
import blast
import blast.blUtils as vutils
from pygeo import DVGeometryCST
from scipy.optimize import least_squares
import fwk

# Solver openMDAO object
class BlasterSolver(om.ExplicitComponent):
    """OpenMDAO component for the Blaster solver

    Parameters
    ----------
    - cfg {dict} -- Configuration dictionary for the VII coupling
    - icfg {dict} -- Configuration dictionary for the inviscid solver interface
    - iSolverName {str} -- Inviscid solver name (default: 'DART')

    Inputs
    ------
    - aoa {float} -- Angle of attack
    - x_aero {np.array} -- Mesh coordinates

    Outputs
    -------
    - cl {float} -- Lift coefficient
    - cd {float} -- Drag coefficient
    """

    def __init__(self, cfg, icfg, iSolverName='DART', saveAll=False):
        super().__init__()

        self._saveAll = saveAll
        self.nCpt = 0
        self.tms = fwk.Timers()

        if iSolverName == 'DART':
            from blast.interfaces.dart.blDartInterface import DartInterface as interface
        else:
            raise RuntimeError('Solver '+iSolverName+' currently not implemented')

        for i in range(len(cfg['xtrF'])):
            if cfg['xtrF'][i] is None:
                cfg['xtrF'][i] = -1
        self.vsol = blast.Driver(cfg['Re'], cfg['Minf'], cfg['CFL0'], cfg['nSections'], xtrF_top=cfg['xtrF'][0], xtrF_bot=cfg['xtrF'][1], _span=cfg['span'], _verbose=cfg['Verb'])

        self.isol = interface(icfg, self.vsol, cfg)

        # Coupler
        import blast.blCoupler as blastCoupler
        self.coupler = blastCoupler.Coupler(self.isol, self.vsol, _maxCouplIter=cfg['couplIter'], _couplTol=cfg['couplTol'], _iterPrint=cfg['iterPrint'], _resetInv=cfg['resetInv'])

        # Adjoint
        self.adjointSolver = blast.CoupledAdjoint(self.isol.adjointSolver, self.vsol)

        # Single coupler run to initialize viscous sections (which depend on size of upper/lower sides)
        self.coupler.run(write=False)
        self.__write(sfx='baseline')
        self.coupler.reset()
        print('object created')
    
    def setup(self):
        self.add_input('aoa', val=0.0, desc='Flow angle of attack')
        self.add_input('x_aero', val=np.zeros(self.isol.iobj['bnd'].nodes.size() * self.isol.getnDim()), desc='Surface mesh coordinates')
        self.add_input('visc_mode', val=True) # Inviscid or viscous mode
        self.add_output('cl', val=0.0, desc='Lift coefficient')
        self.add_output('cd', val=0.0, desc='Drag coefficient')
    
    def setup_partials(self):
        self.declare_partials('cl', 'aoa', method='exact')
        self.declare_partials('cd', 'aoa', method='exact')
        self.declare_partials('cl', 'x_aero', method='exact')
        self.declare_partials('cd', 'x_aero', method='exact')


    def compute_partials(self, inputs, partials, discrete_inputs=None):
        """Compute partial derivatives using the adjoint solver
        """
        nDim = self.isol.getnDim()
        self.tms['adj'].start()

        print(ccolors.ANSI_BLUE, 'Computing derivatives in mode '+('viscous' if inputs['visc_mode'][0] else 'inviscid'), ccolors.ANSI_RESET)
        
        # Run inviscid adjoint solver
        self.isol.iobj['adj'].run()

        dClaoa = 0.
        dCdaoa = 0.
        dClx = np.zeros(self.isol.iobj['bnd'].nodes.size() * nDim)
        dCdx = np.zeros(self.isol.iobj['bnd'].nodes.size() * nDim)

        if inputs['visc_mode'][0]:
            self.adjointSolver.reset()
            self.adjointSolver.run()
            dClaoa = self.adjointSolver.tdCl_AoA
            dCdaoa = self.adjointSolver.tdCd_AoA

            # Mesh derivatives
            for inod, n in enumerate(self.isol.iobj['bnd'].nodes):
                for idim in range(nDim):
                    dClx[inod*nDim + idim] = self.adjointSolver.tdCl_x[n.row][idim]
                    dCdx[inod*nDim + idim] = self.adjointSolver.tdCd_x[n.row][idim]
        elif not inputs['visc_mode'][0]:
            dClaoa = self.isol.iobj['adj'].dClAoa
            dCdaoa = self.isol.iobj['adj'].dCdAoa

            # Mesh derivatives
            for inod, n in enumerate(self.isol.iobj['bnd'].nodes):
                for idim in range(nDim):
                    dClx[inod*nDim + idim] = self.isol.iobj['adj'].dClMsh[n.row][idim]
                    dCdx[inod*nDim + idim] = self.isol.iobj['adj'].dCdMsh[n.row][idim]
        else:
            raise ValueError('omBlasterSolver::compute_partials - Invalid mode', inputs['visc_mode'][0])

        # AoA derivatives
        partials['cl', 'aoa'] = dClaoa * np.pi/180 # /°
        partials['cd', 'aoa'] = dCdaoa * np.pi/180 # /°
        partials['cl', 'x_aero'][0] = dClx
        partials['cd', 'x_aero'][0] = dCdx
        self.tms['adj'].stop()

    def compute(self, inputs, outputs, discrete_inputs=None):
        """Update the mesh, aoa and run the solver
        """
        self.tms['update'].start()
        nDim = self.isol.getnDim()

        # Modify surface and deform mesh
        self.isol.iobj['mrf'].savePos()
        for inod, n in enumerate(self.isol.iobj['bnd'].nodes):
            for idim in range(nDim):
                n.pos[idim] = inputs['x_aero'][inod*nDim + idim]
        self.isol.iobj['mrf'].deform()
        self.isol.updateMesh()

        # Update angle of attack
        aoa = inputs['aoa'][0]
        self.isol.setAoA(aoa*np.pi/180) # aoa is given in deg
        self.tms['update'].stop()

        # Run forward problem
        self.tms['fwd'].start()
        print(ccolors.ANSI_BLUE, 'Computing flow for AoA=', round(aoa,2), 'in mode '+('viscous' if inputs['visc_mode'][0] else 'inviscid'), ccolors.ANSI_RESET)

        if inputs['visc_mode'][0]:
            self.coupler.reset()
            self.coupler.run(write=False)
        elif not inputs['visc_mode'][0]:
            self.isol.reset()
            self.isol.run()
        else:
            raise ValueError('omBlasterSolver::compute - Invalid mode', inputs['visc_mode'][0])
        self.tms['fwd'].stop()

        # write solutions
        self.__write(str(self.nCpt) if self._saveAll else 'opt')

        outputs['cl'] = self.isol.getCl()
        outputs['cd'] = self.isol.getCd()
        if inputs['visc_mode'][0]:
            outputs['cd'] += self.vsol.Cdf # Add viscous contribution
        self.nCpt += 1

    def __write(self, sfx='0'):
        sfx = '_'+sfx
        vutils.getSolution(self.isol.sec, write=True, toW=['x', 'y', 'cf'], sfx=sfx) # Boundary layer
        self.isol.writeCp(sfx=sfx) # Cp
        self.isol.iobj['wrt'].save(self.isol.iobj['pbl'].msh.name + sfx) # Mesh
        self.isol.save(sfx) # Solution
        np.savetxt('shape'+sfx+'.txt', self.isol.vBnd[0][0].nodesCoord) # Shape
        np.savetxt('loads'+sfx+'.txt', [self.isol.getCl(), self.isol.getCd() + self.vsol.Cdf, self.isol.getCm()]) # Loads

    def getBoundaryMesh(self):
        """Initial surface mesh coordinates
        """
        return BlasterMesh(nDim=self.isol.getnDim(), bnd=self.isol.iobj['bnd'])
    
    def getBoundaryNACAGeometry(self):
        """Airfoil naca-4digits geometry
        """
        return BlasterNACAGeometry(nDim=self.isol.getnDim(), nNodes=self.isol.iobj['bnd'].nodes.size(), nodeRows=self.isol.vBnd[0][0].connectListNodes)
    
    def getBoundaryFFDGeometry(self):
        """Airfoil FFD geometry
        """
        with open('surface_DVGeo.dat', 'w') as f:
            for i, node_coord in enumerate(self.isol.vBnd[0][0].nodesCoord):
                f.write(f"{node_coord[0]} {node_coord[1]}\n")
        
        _dvGeo = DVGeometryCST('surface_DVGeo.dat', numCST=4)
        _dvGeo.addDV("upper_shape", dvType="upper", lowerBound=-0.1, upperBound=0.5)
        _dvGeo.addDV("lower_shape", dvType="lower", lowerBound=-0.5, upperBound=0.1)
        _dvGeo.addDV("class_shape_n1", dvType="N1")
        _dvGeo.addDV("class_shape_n2", dvType="N2")
        _dvGeo.addDV("chord", dvType="chord")
        points = np.array(np.loadtxt('surface_DVGeo.dat'))
        points = np.hstack((points, np.zeros((points.shape[0], 1))))  # add 0s for z coordinates (unused)
        _dvGeo.addPointSet(points, "airfoil_coords")
        _dvGeo.update("airfoil_coords")
        return BlasterFFDGeometry(nDim=self.isol.getnDim(), nNodes=self.isol.iobj['bnd'].nodes.size(), nodeRows=self.isol.vBnd[0][0].connectListNodes, dvGeo=_dvGeo)

# Updater to activate viscous mode
class BlasterModeUpdater(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('tolerance', default=1e-6, types=float, desc='Tolerance value to trigger visc_mode')
        self.options.declare('force_mode', default=None, types=str, desc='Force ''visc'' mode or ''inviscid'' mode')

    def setup(self):
        self.add_input('current_objective', val=1.0)
        self.add_output('visc_mode', val=0.0)
        self._previous_objective = 0.0
        self._lock = False

    def compute(self, inputs, outputs):

        # Forced mode case
        if self.options['force_mode'] is not None:
            if self.options['force_mode'] == 'visc':
                outputs['visc_mode'] = True
            elif self.options['force_mode'] == 'inv':
                outputs['visc_mode'] = False
            else:
                raise RuntimeError('Invalid force_mode', self.options['force_mode'])
            return

        tolerance = self.options['tolerance']
        relDiff = abs((inputs['current_objective']-self._previous_objective)/inputs['current_objective'])
        print('curr', inputs['current_objective'][0], 'prev', self._previous_objective)
        print('relDiff', relDiff, 'tolerance', tolerance)
        if relDiff <= tolerance and relDiff != 0. or self._lock:
            outputs['visc_mode'] = True
            self._lock = True
        else:
            outputs['visc_mode'] = False
        print('output visc', outputs['visc_mode'])
        self._previous_objective = inputs['current_objective'][0]
    
    def compute_partials(self, inputs, partials):
        pass

# Surface mesh
class BlasterMesh(om.IndepVarComp):
    """Initial 2D surface mesh coordinates
    
    Parameters
    ----------
    - nDim {int} -- Number of dimensions
    - bnd {object} -- Boundary object containing the mesh

    Outputs
    -------
    - x_aero0 {np.array} -- Initial mesh coordinates
    """

    def initialize(self):
        self.options.declare('nDim')
        self.options.declare('bnd', recordable=False)
    
    def setup(self):
        bnd = self.options['bnd']
        nDim = self.options['nDim']
        x_aero0 = np.zeros(bnd.nodes.size() * nDim)
        for inod, n in enumerate(bnd.nodes):
            for idim in range(nDim):
                x_aero0[inod*nDim + idim] = n.pos[idim]
        self.add_output('x_aero0', val=x_aero0)

class BlasterNACAGeometry(om.ExplicitComponent):
    """Airfoil parametrization as a NACA 4-digit airfoil.
    Uses the NACA 4-digit airfoil parametrization to generate an airfoil from a given x
    distribution.
    Parameters become tau (thickness), eps (max camber) and p (max camber position).

    Parameters
    ----------
    - nDim {int} -- Number of dimensions
    - nNodes {int} -- Number of nodes
    - nodeRows {np.array} -- Nodal indexes

    Inputs
    ------
    - x_aero0 {np.array} -- Initial mesh
    - tau {float} -- Airfoil thickness
    - eps {float} -- Airfoil max camber
    - p {float} -- Airfoil max camber position

    Outputs
    -------
    - x_aero_out {np.array} -- Airfoil coordinates
    """

    def initialize(self):
        self.options.declare('nDim')
        self.options.declare('nNodes')
        self.options.declare('nodeRows')

    def setup(self):
        nNodes = self.options['nNodes']
        nDim = self.options['nDim']
        self.add_input('x_aero0', np.zeros(nNodes*nDim)) # Initial mesh
        self.add_input('tau', val = 0.12, desc='Airfoil thickness')
        self.add_input('eps', val = 0., desc='Airfoil max camber')
        self.add_input('p', val = 0., desc='Airfoil max camber position')
        self.add_output('x_aero_out', val = np.zeros(nNodes*nDim))

    def setup_partials(self):
        self.declare_partials('x_aero_out', 'tau', method='fd')
        self.declare_partials('x_aero_out', 'eps', method='fd')
        self.declare_partials('x_aero_out', 'p', method='fd')
    
    def compute(self, inputs, outputs):
        """Compute airfoil coordinates
        Points are sorted in seilig format to identify upper and lower surfaces.
        Then the NACA 4-digit airfoil is generated and the points are sorted back
        to the original mesh format."""

        x_aero0 = inputs['x_aero0']
        tau = inputs['tau'][0]
        eps = inputs['eps'][0]
        p = inputs['p'][0]
        nDim = self.options['nDim']

        # Retreive selig coordinates
        x_seilig = np.zeros(len(self.options['nodeRows']))
        for i, val in enumerate(self.options['nodeRows']):
            x_seilig[i] = x_aero0[val*nDim]
        # Generate airfoil
        # Format tau eps and p with 2 digits after coma
        print(f"tau: {tau:.10f}, eps: {eps:.10f}, p: {p:.10f}")
        xb, yb = self.__generateAirfoil(x_seilig, tau, eps, p)
        # Go back to normal mesh
        x_aero_out = np.zeros(len(x_aero0))
        for i, val in enumerate(self.options['nodeRows']):
            x_aero_out[val*nDim] = xb[i]
            x_aero_out[val*nDim+1] = yb[i]
        outputs['x_aero_out'] = x_aero_out
    
    def __generateAirfoil(self, x, tau, eps, p):
        """Generate NACA 4-digit airfoil from x coordinates points
        
        Arguments
        ---------
        - x {np.array} -- x coordinates sorted in seilig format
        - tau {float} -- Airfoil thickness
        - eps {float} -- Airfoil max camber
        - p {float} -- Airfoil max camber position
        
        Returns
        -------
        - tuple -- x and y coordinates
        """

        T = 10 * tau * (
          0.2969 * np.sqrt(x) 
        - 0.1260 * (x) 
        - 0.3537 * (x)**2 
        + 0.2843 * (x)**3 
        - 0.1015 * (x)**4) # Thickness

        YBar     = np.zeros(len(T))
        dYBar_dx = np.zeros(len(T))
        Y        = np.zeros(len(T))
        xB       = np.zeros(len(T))
        yB       = np.zeros(len(T))

        for i, xi in enumerate(x):
            # Camber
            if xi < p:
                YBar[i]     = ((eps * xi)/(p**2)) * (2 * p - xi)
                dYBar_dx[i] = (2 * eps)/(p**2) * (p - xi)
            elif xi >= p:
                YBar[i]     = (eps * (1 - xi)/((1-p)**2)) * (1 + xi - 2 * p)
                dYBar_dx[i] = (2 * eps)/((1 - p)**2) * (p - xi)

            # Final x & y coordinates
            if i >= len(x)/2 - 1: 
                # Lower surface
                yB[i] = YBar[i] - 0.5 * T[i] * np.cos(np.arctan(dYBar_dx[i]))
                xB[i] = x[i]    + 0.5 * T[i] * np.sin(np.arctan(dYBar_dx[i]))
            else: 
                # Upper surface
                yB[i] = YBar[i] + 0.5 * T[i] * np.cos(np.arctan(dYBar_dx[i]))
                xB[i] = x[i]    - 0.5 * T[i] * np.sin(np.arctan(dYBar_dx[i]))    
        return (xB, yB)

class BlasterFFDGeometry(om.ExplicitComponent):
    """Airfoil parametrization as a FFD airfoil.
    Uses the FFD airfoil parametrization to generate an airfoil from a given x
    distribution.
    """

    def initialize(self):
        self.options.declare('nDim', desc='Number of dimensions')
        self.options.declare('nNodes', desc='Number of surface nodes')
        self.options.declare('nodeRows', desc='Nodal index of the surface nodes sorted in seilig format')
        self.options.declare('dvGeo', desc='Parametrization of the geometry', recordable=False)

    def setup(self):
        nNodes = self.options['nNodes']
        nDim = self.options['nDim']

        self.add_input('x_aero0', np.zeros(nNodes*nDim)) # Initial mesh
        self.add_input('upper_shape', [0.17553542, 0.14723497, 0.14531762, 0.13826955])
        self.add_input('lower_shape', [-0.16935506, -0.15405552, -0.13912672, -0.14168947])
        self.add_input('class_shape_n1', 0.5)
        self.add_input('class_shape_n2', 1.0)
        self.add_output('x_aero_out', val = np.zeros(nNodes*nDim))
        self.add_output('thickness', val = .12)
        self.add_output('LEradius', val = 0.02)

    def setup_partials(self):
        self.declare_partials('x_aero_out', 'upper_shape', method='fd')
        self.declare_partials('x_aero_out', 'lower_shape', method='fd')
        #self.declare_partials('x_aero_out', 'class_shape_n1', method='fd')
        #self.declare_partials('x_aero_out', 'class_shape_n2', method='fd')
        self.declare_partials('thickness', 'upper_shape', method='fd')
        self.declare_partials('thickness', 'lower_shape', method='fd')
        # self.declare_partials('thickness', 'class_shape_n1', method='fd')
        # self.declare_partials('thickness', 'class_shape_n2', method='fd')
        self.declare_partials('LEradius', 'upper_shape', method='fd')
        self.declare_partials('LEradius', 'lower_shape', method='fd')
        # self.declare_partials('LEradius', 'class_shape_n1', method='fd')
        # self.declare_partials('LEradius', 'class_shape_n2', method='fd')

    def compute(self, inputs, outputs):
        nDim = self.options['nDim']

        self.options['dvGeo'].setDesignVars({"upper_shape": inputs['upper_shape'],
                             "lower_shape": inputs['lower_shape'],
                             "class_shape_n1": inputs['class_shape_n1'],
                             "class_shape_n2": inputs['class_shape_n2']})
        points1 = self.options['dvGeo'].update("airfoil_coords")

        # Compute thickness at 1/4 chord
        upper_side = points1[:len(points1)//2]
        lower_side = points1[len(points1)//2:]
        arg_up = np.argmin(np.abs(upper_side[:, 0] - 0.38))
        arg_low = np.argmin(np.abs(lower_side[:, 0] - 0.38))
        maxThickness = upper_side[arg_up, 1] - lower_side[arg_low, 1]

        # Extract leading edge points (assuming points are sorted by x)
        le_points = points1[np.abs(points1[:, 0]) < 0.01]  # Adjust threshold as needed

        # Separate upper and lower surface points
        upper_surface = le_points[le_points[:, 1] >= 0]
        lower_surface = le_points[le_points[:, 1] < 0]

        # Combine upper and lower surface points for circle fitting
        le_points_combined = np.vstack((upper_surface, lower_surface))

        # Fit circle to leading edge points
        center, radius = self.fit_circle(le_points_combined[:, 0], le_points_combined[:, 1])

        # Go back to normal mesh
        x_aero_out = np.zeros(len(inputs['x_aero0']))
        for i, val in enumerate(self.options['nodeRows']):
            for idim in range(nDim):
                x_aero_out[val*nDim+idim] = points1[i, idim]
        outputs['x_aero_out'] = x_aero_out
        outputs['thickness'] = maxThickness
        outputs['LEradius'] = radius

    def fit_circle(self, x, y):
        def calc_R(xc, yc):
            return np.sqrt((x - xc)**2 + (y - yc)**2)

        def f_2(c):
            Ri = calc_R(*c)
            return Ri - Ri.mean()

        x_m = np.mean(x)
        y_m = np.mean(y)
        center_estimate = x_m, y_m
        result = least_squares(f_2, center_estimate)
        center = result.x
        Ri = calc_R(*center)
        R = Ri.mean()
        return center, R
